
/*
**	This file contains routines for performing edge detection on
**	greyscale images.
*/

#include <stdio.h>
#include <math.h>
#include <windows.h>
#include "globals.h"

	/*
	**	This routine performs edge detection using a window with
	**	Sobel weights.  It outpus both the magnitude and direction
	**	image maps.
	*/

void SobelEdgeDetect(unsigned char *InputImage,
					 unsigned char *MagOutputImage,
					 unsigned char *DirOutputImage,
					 int nrows,int ncols,int nbytes)

{
int			b,r,c,*TempArray1,*TempArray2,strength;
double		angle;
//char		text[200];
//HDC			hDC;

TempArray1=(int *)calloc((nbytes*nrows*ncols),sizeof(int));
TempArray2=(int *)calloc((nbytes*nrows*ncols),sizeof(int));
for (b=0; b<nbytes; b++)
  for (r=1; r<(nrows-1); r++)
    {
/*
	sprintf(text,"%d%% complete",(int)(100.0*(double)(b*nrows+r)/
				(double)(nbytes*nrows)));
	hDC=GetDC(MainWnd);
    TextOut(hDC,0,0,text,strlen(text));
    ReleaseDC(MainWnd,hDC);
*/
    for (c=1; c<(ncols-1); c++)
      {
	  TempArray1[nbytes*(r*ncols+c)+b]=InputImage[nbytes*((r-1)*ncols+c+1)+b]
						+InputImage[nbytes*((r+1)*ncols+c+1)+b]
						-InputImage[nbytes*((r-1)*ncols+c-1)+b]
						-InputImage[nbytes*((r+1)*ncols+c-1)+b]
						+(2*(InputImage[nbytes*(r*ncols+c+1)+b]
						-InputImage[nbytes*(r*ncols+c-1)+b]));
	  TempArray2[nbytes*((r*ncols)+c)+b]=InputImage[nbytes*((r+1)*ncols+c+1)+b]
						+InputImage[nbytes*((r+1)*ncols+c-1)+b]
						-InputImage[nbytes*((r-1)*ncols+c+1)+b]
						-InputImage[nbytes*((r-1)*ncols+c-1)+b]
						+(2*(InputImage[nbytes*((r+1)*ncols+c)+b]
						-InputImage[nbytes*((r-1)*ncols+c)+b]));
	  }
    }
for (r=1; r<(nrows-1); r++)
  {
  for (c=1; c<(ncols-1); c++)
    {
	strength=0;
	for (b=0; b<nbytes; b++)
      strength+=(int)sqrt((double)(SQR(TempArray1[nbytes*(r*ncols+c)+b])+
			SQR(TempArray2[nbytes*(r*ncols+c)+b])));
	strength/=nbytes;
    if (strength >= 254)
      MagOutputImage[r*ncols+c]=255;
    else
      MagOutputImage[r*ncols+c]=(unsigned char)strength;
	angle=0.0;
	for (b=0; b<nbytes; b++)
      if (TempArray1[nbytes*(r*ncols+c)+b] == 0)
        angle+=3.1415927/2.0;
      else
        angle+=atan((double)TempArray2[nbytes*(r*ncols+c)+b]/(double)TempArray1[nbytes*(r*ncols+c)+b]);
	if (DirOutputImage != NULL)
      DirOutputImage[r*ncols+c]=(unsigned char)((angle+(3.1415927/2.0))*255.0/3.1415927);
    }
  }
	/* Fill in 1-pixel width border with zeroes (window doesn't reach) */
for (r=0; r<nrows; r++)
  {
  MagOutputImage[r*ncols]=MagOutputImage[r*ncols+ncols-1]=0;
  if (DirOutputImage != NULL)
    DirOutputImage[r*ncols]=DirOutputImage[r*ncols+ncols-1]=0;
  }
for (c=0; c<ncols; c++)
  {
  MagOutputImage[c]=MagOutputImage[nrows*ncols-c-1]=0;
  if (DirOutputImage != NULL)
    DirOutputImage[c]=DirOutputImage[nrows*ncols-c-1]=0;
  }
free(TempArray1);
free(TempArray2);
}


	/*
	** Reduces a binary image (zeros and 255s) to its skeleton,
	** preserving original connectivity
	*/

void ThinEdgeImage(unsigned char *InputImage,
				   unsigned char *OutputImage,
				   int nrows,int ncols)

{
int				r,c,r2,c2,killed,pass,Transitions,EdgeNeighbors,ThinningFlag;
unsigned char	*TempArray;

TempArray=(unsigned char *)calloc(nrows*ncols,sizeof(unsigned char));
for (r=0; r<nrows*ncols; r++)
  OutputImage[r]=InputImage[r];
pass=0;
do
  {
  for (r=0; r<nrows*ncols; r++)
    TempArray[r]=OutputImage[r];
  ThinningFlag=0;
  pass++;
  killed=0;
  for (r=2; r<(nrows-2); r++)
    {
    for (c=2; c<(ncols-2); c++)
      {
      if (TempArray[r*ncols+c] == 255)
		{
		Transitions=0;
		if (TempArray[(r-1)*ncols+c] == 0  &&  TempArray[(r-1)*ncols+c+1] == 255)
		  Transitions++;
		if (TempArray[(r-1)*ncols+c+1] == 0  &&  TempArray[r*ncols+c+1] == 255)
		  Transitions++;
		if (TempArray[r*ncols+c+1] == 0  &&  TempArray[(r+1)*ncols+c+1] == 255)
		  Transitions++;
		if (TempArray[(r+1)*ncols+c+1] == 0  &&  TempArray[(r+1)*ncols+c] == 255)
		  Transitions++;
		if (TempArray[(r+1)*ncols+c] == 0  &&  TempArray[(r+1)*ncols+c-1] == 255)
		  Transitions++;
		if (TempArray[(r+1)*ncols+c-1] == 0  &&  TempArray[r*ncols+c-1] == 255)
		  Transitions++;
		if (TempArray[r*ncols+c-1] == 0  &&  TempArray[(r-1)*ncols+c-1] == 255)
		  Transitions++;
		if (TempArray[(r-1)*ncols+c-1] == 0  &&  TempArray[(r-1)*ncols+c] == 255)
		  Transitions++;
		if (Transitions != 1)
		  continue;	/* can only have one edge to non-edge transition */
		EdgeNeighbors=0;
		for (r2=-1; r2<=1; r2++)
		  for (c2=-1; c2<=1; c2++)
		    if (TempArray[(r+r2)*ncols+c+c2] == 255)
			  EdgeNeighbors++;
		if (EdgeNeighbors < 3  ||  EdgeNeighbors > 7)
		  continue;	/* must have 3 to 7 edge pixel neighbors */
		if (TempArray[((r-1)*ncols)+c] == 0  || 
				TempArray[(r*ncols)+c+1] == 0  ||
				(TempArray[(r*ncols)+c-1] == 0  &&
				TempArray[((r+1)*ncols)+c] == 0))
		  {
		  ThinningFlag=1;
		  killed++;
		  OutputImage[(r*ncols)+c]=0;
		  }
		}
      }
    }
  }
while (ThinningFlag == 1);
free(TempArray);
}


