
	/*******************************************************************
	** Routines in this file convert geometric shapes (lines and circles)
	** into sets of pixel coordinates.
	*******************************************************************/

#include <stdio.h>
#include <math.h>
#include <windows.h>
#include "globals.h"

	/*******************************************************************
	** Given two endpoints of a line segment (integers), this routine
	** computes the indices of the integer points of the segment.
	*******************************************************************/

void MakeLineIndices(int x1,int y1,int x2,int y2,	/* inputs -- endpoints of line segment */
					 int ROWS,int COLS,				/* inputs -- clipping width & height */
					 int *indices,					/* output -- indices of pixels on line */
					 int *tp)						/* output -- total pixels in line segment */

{
float   x,y;
int     temp,swap,i;

if (x1 > x2)
  { temp=x1; x1=x2; x2=temp; temp=y1; y1=y2; y2=temp; swap=1; }
else
  swap=0;
x=(float)x1; y=(float)y1;
*tp=0;
do
  {
  if ((int)x >= 0  &&  (int)x < COLS  &&  (int)y >= 0  &&  (int)y < ROWS)
    {
    indices[*tp]=(int)y*COLS+(int)x;
    (*tp)++;
    }
  if (x1 == x2  &&  y1 == y2)
    break;
  if (abs(x2-x1) > abs(y2-y1))
    {
    x+=1.0;
    y+=(float)(y2-y1)/(float)(x2-x1);
    }
  else
    {
    if (y2 > y1)
      y+=1.0;
    else
      y-=1.0;
    x+=(float)(x2-x1)/(float)fabs((double)(y2-y1));
    }
  }
while (fabs((double)x-(double)x2) > 0.5  ||  fabs((double)y-(double)y2) > 0.5);
if (swap)
  for (i=0; i<(*tp)/2; i++)
    { temp=indices[i]; indices[i]=indices[(*tp)-1-i]; indices[(*tp)-1-i]=temp; }
}








	/*******************************************************************
	** Given the center and radius of a circle (real #s), this routine
	** computes the indices of the integer points on the circle.
	*******************************************************************/

void MakeCircleIndices(double cx,double cy,double ra, /* inputs -- center and radius of circle */
					   int ROWS,int COLS,				/* inputs -- clipping width & height */
					   int *indices,				/* output -- indices of points on circle */
					   int *tp)					/* output -- total points in circle */

{
double	x,y;
int	pass;	/* 1 => building top half of circle; 2 => bottom half */

if (ra < 1.0)
  {
  *tp=0;
  return;
  }
x=0.0-ra; y=0.0;
pass=1;			/* build top half of circle left->right */
*tp=0;
while (pass < 3)
  {
  if ((int)(cx+x) >= 0  &&  (int)(cx+x) < COLS  &&
	(int)(cy-y) >= 0  &&  (int)(cy-y) < ROWS)
    {
    indices[*tp]=(int)(cy-y)*COLS+(int)(cx+x);
    (*tp)++;
    }
  if (atan2(fabs(y),fabs(x)) > M_PI/4.0)
    {
    if (pass == 1)
      {
      x+=1.0;
      y=sqrt(ra*ra-x*x);
      }
    else
      {
      x-=1.0;
      y=-sqrt(ra*ra-x*x);
      }
    }
  else if (x < 0.0)
    {
    y+=1.0;
    x=-sqrt(ra*ra-y*y);
    }
  else
    {
    y-=1.0;
    x=sqrt(ra*ra-y*y);
    }
  if (pass == 1  &&  y <= 0.0)
    pass=2;		/*build bottom half of circle right->left */
  else if (pass == 2  &&  y >= 0.0)
    pass=3;
  }
(*tp)--;	/* last point is duplicate of first point */
}




	/*******************************************************************
	** Given two endpoints and a mid-point of a line segment (integers),
	** this routine computs the indices of the integer points along
	** the 2nd-order curve passing through the points.
	*******************************************************************/

void MakeCurveIndices(x1,y1,x2,y2,x3,y3,ROWS,COLS,Extension,indices,tp)

int		x1,y1,x2,y2,x3,y3; /* inputs -- points on segment */
int		ROWS,COLS;	/* inputs -- clipping width & height */
int		Extension;	/* input -- percent to extend segment */
int		*indices;	/* output -- indices of pixels on curve */
int		*tp;		/* output -- total pixels in curve */

{
int     x,y;
double	**a,**b,**dmatrix(),Ax,Bx,Cx,Ay,By,Cy;
double	t,dt;

/*
if (x1 > x2)
  { temp=x1; x1=x2; x2=temp; temp=y1; y1=y2; y2=temp; }
if (x2 > x3)
  { temp=x2; x2=x3; x3=temp; temp=y2; y2=y3; y3=temp; }
if (x1 > x2)
  { temp=x1; x1=x2; x2=temp; temp=y1; y1=y2; y2=temp; }
	/* first solve for the parametric equation coefficients */
a=dmatrix(1,3,1,3);
b=dmatrix(1,3,1,1);
a[1][1]=SQR(0.0); a[1][2]=0.0; a[1][3]=1.0; b[1][1]=x1;
a[2][1]=SQR(0.5); a[2][2]=0.5; a[2][3]=1.0; b[2][1]=x2;
a[3][1]=SQR(1.0); a[3][2]=1.0; a[3][3]=1.0; b[3][1]=x3;
if (gaussj(a,3,b,1) != 1)
  { printf("num recipes error solving for parametric curve\n"); exit(0); }
Ax=b[1][1]; Bx=b[2][1]; Cx=b[3][1];
a[1][1]=SQR(0.0); a[1][2]=0.0; a[1][3]=1.0; b[1][1]=y1;
a[2][1]=SQR(0.5); a[2][2]=0.5; a[2][3]=1.0; b[2][1]=y2;
a[3][1]=SQR(1.0); a[3][2]=1.0; a[3][3]=1.0; b[3][1]=y3;
if (gaussj(a,3,b,1) != 1)
  { printf("num recipes error solving for parametric curve\n"); exit(0); }
Ay=b[1][1]; By=b[2][1]; Cy=b[3][1];
free_dmatrix(a,1,3,1,3);
free_dmatrix(b,1,3,1,1);
	/* estimate delta-t based upon distance between endpoints */
dt=(1.0/sqrt((double)(SQR(x1-x3)+SQR(y1-y3))))/2.0;
	/* find points on parametric curve */
(*tp)=0;
for (t=0-(double)Extension/100.0; t<=1.0+(double)Extension/100.0; t+=dt)
  {
  x=(int)(Ax*SQR(t)+Bx*t+Cx);
  y=(int)(Ay*SQR(t)+By*t+Cy);
  if (y < 0  ||  y >= ROWS  ||  x < 0  ||  x >= COLS)
    continue;
  if (*tp == 0  ||  !(y*COLS+x == indices[(*tp)-1]))
    {
    indices[*tp]=y*COLS+x;
    (*tp)++;
    }
  }
}


