
#include <stdio.h>
#include <math.h>
#include <windows.h>
#include "globals.h"

		/*********************************************************
		** This is an implementation of ``Detection of Blood Vessels
		** in Retinal Images Using Two-Dimensional Matched Filters'',
		** Chadhuri et.al., in IEEE Transactions on Medical Imaging,
		** Sep. '89, pp. 263-269.
		**
		** Coded by Adam Hoover in June '97.
		**
		** Extended to incorporate sub-masks for junction-preserving.
		** Coded by Adam Hoover in July '98.
		**
		** Ported to windows in September '98.
		** Rather than read the filters file, this version computes
		** them using the published Gaussian profile formula.
		*********************************************************/

int Filters[12][15*15];
int TotalFilters=12;
int FilterWidth=15;
int FilterHeight=15;

void MatchSpatialFilterVessels(unsigned char *RawImage,
							   int RawROWS,int RawCOLS,
							   unsigned char *VesselsImage)

{
char			text[200];
int				*FilterImage;
unsigned char	*GreenImage,*MaskImage;
int				f,fi,i,r,c,r2,c2;
int				Score,HighestScore,GlobalLow,GlobalHigh;
HDC				hDC;

		/*********************************************************
		** Create single-band Green image for faster convultion.
		*********************************************************/

if ((GreenImage=(unsigned char *)calloc(1,RawROWS*RawCOLS)) == NULL  ||
	(MaskImage=(unsigned char *)calloc(1,RawROWS*RawCOLS)) == NULL  ||
	(FilterImage=(int *)calloc(sizeof(int),RawROWS*RawCOLS)) == NULL)
  {
  MessageBox(MainWnd,"Error -- out of memory","Crash",MB_OK | MB_APPLMODAL);
  return;
  }
for (i=0; i<RawROWS*RawCOLS; i++)
  GreenImage[i]=RawImage[i*3+1];

		/*********************************************************
		** Read filters from file.
		*********************************************************/
/*
if ((fpt=fopen(FiltersFilename,"r")) == NULL)
  {
  MessageBox(MainWnd,"Unable to open filters file.","Error",MB_OK | MB_APPLMODAL);
  return;
  }
if ((fscanf(fpt,"%d %d %d %d %d",&TotalFilters,&FilterHeight,&FilterWidth,
	&LowFilterValue,&HighFilterValue)) != 5)
  {
  MessageBox(MainWnd,"Header should contain #filters, height, width, low-val, high-val.","Error",MB_OK | MB_APPLMODAL);
  return;
  }
Filters=(int **)calloc(TotalFilters,sizeof(int *));
for (f=0; f<TotalFilters; f++)
  if ((Filters[f]=(int *)calloc(FilterWidth*FilterHeight,sizeof(int))) == NULL)
    {
    MessageBox(MainWnd,"Unable to allocate memory for filters.","Error",MB_OK | MB_APPLMODAL);
	return;
    }
for (f=0; f<TotalFilters; f++)
  {
  for (i=0; i<FilterWidth*FilterHeight; i++)
    if ((fscanf(fpt,"%d",&Filters[f][i])) != 1)
      {
	  sprintf(text,"Expected %d %dx%d matrices in filters file.",
				TotalFilters,FilterHeight,FilterWidth);
      MessageBox(MainWnd,text,"Error",MB_OK | MB_APPLMODAL);
	  return;
      }
  }
fclose(fpt);
*/

CreateVesselsFilters();
MakeStareMask(MaskTopRow,MaskBottomRow,MaskCircle,RawROWS,RawCOLS,MaskImage);

		/*********************************************************
		** Convolve filters with image; vessels image gets highest
		** scoring filter at each pixel.  Values are stored in integer
		** image on this pass, and will be normalized into 8-bit
		** (unsigned char) image on the next pass.  For this reason,
		** global low & high values are found.
		*********************************************************/

GlobalLow=GlobalHigh=-1;
for (r=0; r<RawROWS-FilterHeight; r++)
  {
  sprintf(text,"%d%% complete",(int)(100.0*(double)r/(double)(RawROWS-FilterHeight)));
  hDC=GetDC(MainWnd);
  TextOut(hDC,0,0,text,strlen(text));
  ReleaseDC(MainWnd,hDC);
  for (c=0; c<RawCOLS-FilterWidth; c++)
    {
    if (!(MaskImage[(r+FilterHeight/2)*RawCOLS+c+FilterWidth/2]))
      {
      FilterImage[(r+FilterHeight/2)*RawCOLS+c+FilterWidth/2]=0;
      continue;
      }
    for (f=0; f<TotalFilters; f++)
      {
      Score=0;
      for (r2=r,fi=0; r2<r+FilterHeight; r2++)
        for (c2=c; c2<c+FilterWidth; c2++,fi++)
          if (Filters[f][fi] != 0  &&  MaskImage[r2*RawCOLS+c2])
	        Score+=(GreenImage[r2*RawCOLS+c2]*Filters[f][fi]);
      if (f == 0  ||  Score > HighestScore)
        HighestScore=Score;
      }
    FilterImage[(r+FilterHeight/2)*RawCOLS+c+FilterWidth/2]=HighestScore;
    if ((GlobalLow == -1  ||  HighestScore < GlobalLow)  &&  HighestScore != 0)
      GlobalLow=HighestScore;
    if ((GlobalHigh == -1  ||  HighestScore > GlobalHigh) && HighestScore != 0)
      GlobalHigh=HighestScore;
    }
  }

		/*********************************************************
		** Scale integer-valued filter-response image to 8-bits,
		** then output the vessels image map.
		*********************************************************/

for (r=0; r<RawROWS; r++)
  for (c=0; c<RawCOLS; c++)
    if (!(MaskImage[r*RawCOLS+c]))
      VesselsImage[r*RawCOLS+c]=0;
    else
      VesselsImage[r*RawCOLS+c]=(unsigned char)((FilterImage[r*RawCOLS+c]-GlobalLow)*255/(GlobalHigh-GlobalLow));

		/*********************************************************
		** Free up memory that was dynamically allocated.
		*********************************************************/

free(GreenImage);
free(FilterImage);
/*
for (f=0; f<TotalFilters; f++)
  free(Filters[f]);
free(Filters);
*/
}


		/*********************************************************
		** Uses the constants and formula in the TMI paper to make filters.
		*********************************************************/

void CreateVesselsFilters()

{
double	theta,K[15*15],mean,U,V;
int		r,c,a,u,v,count;

//Filters=(int **)calloc(12,sizeof(int *));
//for (a=0; a<12; a++)
// if ((Filters[a]=(int *)calloc(15*15,sizeof(int))) == NULL)
//    {
//    MessageBox(MainWnd,"Unable to allocate memory for filters.","Error",MB_OK | MB_APPLMODAL);
//	  return;
//    }

for (a=0; a<12; a++)
  {
  theta=15.0*(double)a*M_PI/180.0;
  mean=0.0;
  count=0;
  for (r=0; r<15; r++)
    {
    for (c=0; c<15; c++)
      {
      u=c-7;
      v=0-(r-7);
      U=((double)u*cos(theta)+(double)v*sin(theta));
      V=(0.0-(double)u*sin(theta)+(double)v*cos(theta));
      if (fabs(U) > 3.0*2.0  ||  fabs(V) > 4.5)
        {
        K[r*15+c]=0.0;
        continue;
        }
      K[r*15+c]=0.0-pow(M_E,((0.0-(double)(U*U))/(2.0*2.0*2.0)));
      mean+=K[r*15+c];
      count++;
      }
    }
  for (r=0; r<15; r++)
    {
    for (c=0; c<15; c++)
      {
      u=c-7;
      v=0-(r-7);
      U=((double)u*cos(theta)+(double)v*sin(theta));
      V=(0.0-(double)u*sin(theta)+(double)v*cos(theta));
      if (fabs(U) > 3.0*2.0  ||  fabs(V) > 4.5)
        {
        Filters[a][r*15+c]=0;
        }
      else
        {
        K[r*15+c]=(K[r*15+c]-mean/(double)count)*10.0;
        if (K[r*15+c] < 0.0)
	      K[r*15+c]-=0.5;
	    else
	      K[r*15+c]+=0.5;
		Filters[a][r*15+c]=(int)(K[r*15+c]);
        }
      }
    }
  }

}


