
	/*
	**	This file contains routines for processing greyscale images
	**	through various filters.
	*/


#include <stdio.h>
#include <math.h>
#include <windows.h>
#include "globals.h"

	/*
	** 	This routine replaces every pixel with the median of the window.
	*/

void MedianFilter(unsigned char *InputImage,
				  unsigned char *OutputImage,
				  int nrows,int ncols,
				  int nbytes,	/* 1 or 3 (bands) */
				  int size) /* 0 means 3x3, 1 means 5x5, 2 means 7x7 */

{
int		b,r,c,r2,c2,s,s2,temp,TotalPoints,sort[121];
char	text[200];
HDC		hDC;

size=size+1;	/* half of filter size, rounded down */
if (size < 1) size=1;	/* 3x3 */
if (size > 5) size=5;	/* 11x11 */
for (b=0; b<nbytes; b++)
  {
  for (r=0; r<nrows; r++)
    {
	sprintf(text,"%d%% complete",(int)(100.0*(double)(b*nrows+r)/
				(double)(nbytes*nrows)));
	hDC=GetDC(MainWnd);
	TextOut(hDC,0,0,text,strlen(text));
    ReleaseDC(MainWnd,hDC);
    for (c=0; c<ncols; c++)
      {
      TotalPoints=0;
      for (r2=-size; r2<=size; r2++)
        for (c2=-size; c2<=size; c2++)
		  if (r+r2 >= 0  &&  r+r2 < nrows  &&  c+c2 >= 0  &&  c+c2 < ncols)
		    {
		    sort[TotalPoints]=InputImage[nbytes*((r+r2)*ncols+c+c2)+b];
		    TotalPoints++;
		    }
      for (s=0; s<=TotalPoints/2; s++)
        for (s2=s+1; s2<TotalPoints; s2++)
		  if (sort[s] > sort[s2])
		    {
		    temp=sort[s];
		    sort[s]=sort[s2];
		    sort[s2]=temp;
		    }
      if (((TotalPoints/2)*2) == TotalPoints)
        OutputImage[nbytes*(r*ncols+c)+b]=(sort[TotalPoints/2]+sort[TotalPoints/2-1])/2;
      else
        OutputImage[nbytes*(r*ncols+c)+b]=sort[TotalPoints/2];
	  }
    }
  }
}



	/*
	**	This routine replaces every pixel with the gaussian average of
	**	the window.
	*/

void GaussianFilter(unsigned char *InputImage,
			   unsigned char *OutputImage,
			   int nrows,int ncols,
			   int nbytes,	/* 1 or 3 (bands) */
			   int size) /* 0 means 0.5 pixels/sigman, 1 means 1.0, 2 means 1.5, etc. */

{
int				b,r,c,r2,c2,total,ww[8];
unsigned char	*TempImage;

switch(size)
  {
  case 4:
    size=7;
    ww[0]=1;	ww[1]=1;	ww[2]=3;	ww[3]=5;
    ww[4]=5;	ww[5]=14;	ww[6]=14;	ww[7]=14;
    break;
  case 3:
    size=6;
    ww[0]=1;	ww[1]=1;	ww[2]=4;	ww[3]=7;
    ww[4]=12;	ww[5]=17;	ww[6]=16;
    break;
  case 2:
    size=4;
    ww[0]=2;	ww[1]=5;	ww[2]=9;	ww[3]=23;
    ww[4]=22;
    break;
  case 1:
    size=3;
    ww[0]=1;	ww[1]=8;	ww[2]=24;	ww[3]=34;
    break;
  case 0:
    size=1;
    ww[0]=16;	ww[1]=68;
    break;
  }
TempImage=(unsigned char *)calloc(nrows*ncols,sizeof(unsigned char));
	/* Window is symmetric, so filtering can be done in two 1D passes */
for (b=0; b<nbytes; b++)
  {
  for (r=size; r<(nrows-size); r++)
	{
	for (c=size; c<(ncols-size); c++)
	  {
	  total=0;
	  for (r2=1; r2<=size; r2++)
		total=total+((InputImage[nbytes*((ncols*(r-r2))+c)+b]+
			  InputImage[nbytes*((ncols*(r+r2))+c)+b])*ww[size-r2]);
	  TempImage[(ncols*r)+c]=(total+(InputImage[nbytes*((ncols*r)+c)+b]*ww[size]))/100;
	  }
	}

  for (r=size; r<(nrows-size); r++)
	{
	for (c=size; c<(ncols-size); c++)
	  if (r < size  ||  r >= (nrows-size)  ||  c < size  ||  c >= (ncols-size))
	  /* Fill in border (outside filtering window's reach) with zeroes */
		OutputImage[nbytes*(r*ncols+c)+b]=0;
	  else
		{
		total=0;
		for (c2=1; c2<=size; c2++)
		  {
		  total=total+((TempImage[(ncols*r)+(c-c2)]+
		  TempImage[(ncols*r)+(c+c2)])*ww[size-c2]);
		  }
		OutputImage[nbytes*((ncols*r)+c)+b]=
		  (total+(TempImage[(ncols*r)+c]*ww[size]))/100;
		}
	}
  }
free(TempImage);
}




	/*
	**	This routine replaces every pixel with the mean average of
	**	the window.
	*/

void MeanFilter(unsigned char *InputImage,
		   unsigned char *OutputImage,
		   int nrows,int ncols,int nbytes,
		   int WindowSize) /* 0 means 3x3, 1 means 5x5, 2 means 7x7 */

{
int				r,c,b,WindowTotal;
unsigned char	*TempArray;
char	text[200];
HDC		hDC;

WindowSize=(WindowSize+1)*2+1;
TempArray=(unsigned char *)calloc(nrows*ncols,sizeof(unsigned char));
for (b=0; b<nbytes; b++)
  {
  for (r=0; r<nrows; r++)
    {
	sprintf(text,"%d%% complete",(int)(100.0*(double)(b*nrows+r)/
				(double)(nbytes*nrows)));
	hDC=GetDC(MainWnd);
	TextOut(hDC,0,0,text,strlen(text));
    ReleaseDC(MainWnd,hDC);
    WindowTotal=0;
    for (c=0; c<WindowSize; c++)
      WindowTotal+=InputImage[(r*ncols+c)*nbytes+b];
    TempArray[r*ncols+(WindowSize/2)]=WindowTotal/WindowSize;
    for (c=0; c<ncols-WindowSize; c++)
      {
      WindowTotal-=InputImage[(r*ncols+c)*nbytes+b];
      WindowTotal+=InputImage[(r*ncols+WindowSize+c)*nbytes+b];
      TempArray[r*ncols+c+1+(WindowSize/2)]=WindowTotal/WindowSize;
      }
    }
  for (c=0; c<ncols; c++)
    {
    WindowTotal=0;
    for (r=0; r<WindowSize; r++)
      WindowTotal+=TempArray[r*ncols+c];
    OutputImage[(WindowSize/2)*ncols+c]=WindowTotal/WindowSize;
    for (r=0; r<nrows-WindowSize; r++)
      {
      WindowTotal-=TempArray[r*ncols+c];
      WindowTotal+=TempArray[(r+WindowSize)*ncols+c];
      OutputImage[((r+1+(WindowSize/2))*ncols+c)*nbytes+b]=WindowTotal/WindowSize;
      }
    }
  }
free(TempArray);
}





	/*
	**	This routine replaces every pixel with the mode of
	**	the window.  The mode is the most commonly occuring value.
	*/

void ModeFilter(unsigned char *InputImage,
				unsigned char *OutputImage,
				int nrows,int ncols,
				int WindowSize) /* 0 means 3x3, 1 means 5x5, 2 means 7x7 */

{
int		i,r,c,r2,c2,mode;
int		Values[49],Count[49],TotalValues;

WindowSize=WindowSize+1;
for (r=0; r<nrows; r++)
  for (c=0; c<ncols; c++)
    {
    TotalValues=0;
    for (r2=r-WindowSize; r2<=r+WindowSize; r2++)
      for (c2=c-WindowSize; c2<=c+WindowSize; c2++)
	if (r2 >= 0  &&  r2 < nrows  &&  c2 >= 0  &&  c2 < ncols)
	  {
	  for (i=0; i<TotalValues; i++)
	    if (InputImage[r2*ncols+c2] == Values[i])
	      break;
	  if (i == TotalValues)
	    {
	    Count[i]=0;
	    Values[i]=InputImage[r2*ncols+c2];
	    TotalValues++;
	    }
	  Count[i]++;
	  }
    mode=0;
    for (i=1; i<TotalValues; i++)
      if (Count[i] > Count[mode])
	mode=i;
    OutputImage[r*ncols+c]=Values[mode];
    }
}




